==> blueprint-client-0.2.18.jar: blueprint/client.clj <==
(ns blueprint.client
  (:require [blueprint.client.url :as url]
            [blueprint.core :as core]
            [blueprint.registry :as reg]
            [blueprint.interceptor :as bpi]
            [blueprint.client.interceptors :as interceptors]
            [aleph.http :as http]
            [muuntaja.core :as m]
            [exoscale.interceptor.manifold :as ixm]
            [blueprint.client.spec-gen :as bsg]
            [clojure.spec.alpha :as s]
            [exoscale.ex :as ex]))

(def default-interceptors
  [interceptors/execute-request
   interceptors/lens-response
   interceptors/decode
   interceptors/parse
   interceptors/get-response])

(defn- build-chain
  "Build an interceptor chain in three phases:

  - Start from the default chain
  - Process additional interceptors provided in `::additional`
  - Removes interceptors by name based on those provided in `::disabled`
  - Call the interceptor build step if any with the provided config"
  [config]
  (ex/assert-spec-valid ::interceptors config)
  (bpi/build-chain default-interceptors config))

(def build-with bpi/build-with)

;;
;; Http Client
;;

(defn- base-url
  "Gets the base url for the defined api."
  [{:keys [servers] :as parsed-api}]
  (if-let [prefix (get-in parsed-api [:blueprint.options :path-prefix])]
    (str (:url (first servers)) prefix)
    (:url (first servers))))

(defn- resolve-server
  "Resolves the correct server according to the configured option."
  [str-or-ifn  opts]
  (let [[t val] (s/conform :client-options/server str-or-ifn)]
    (cond (= t :str) val
          (= t :ifn) (val opts)
          :else (throw (ex-info (str "Unrecognized server option: " str-or-ifn) {:type t
                                                                                 :val  val})))))

(defrecord Client [parsed-api command-defs server connection-pool])

(defn make-client
  "Create a blueprint client from a blueprint api definition."
  ([api-def] (make-client api-def {}))
  ([api-def {:keys [ssl-context server] :as default-options}]
   (ex/assert-spec-valid ::api-definition api-def)
   (ex/assert-spec-valid ::client-options default-options)

   (let [parsed-api      (core/parse api-def)
         command-defs    (bsg/apidef->commands-map parsed-api)
         connection-pool (if ssl-context
                           (http/connection-pool
                             {:connection-options {:ssl-context ssl-context}})
                           http/default-connection-pool)
         server          (or server (base-url parsed-api))]
     (->Client parsed-api command-defs server connection-pool))))

(def default-options
  {:pool-timeout       10e3
   :connection-timeout 10e3
   :request-timeout    10e3
   :read-timeout       10e3})

(defn invoke
  "Invoke a `command` on the remote blueprint-based api.
  Clients are created with `(blueprint.client/make-client api-def)`.
  Supported options in `opts`:

  | key                   | description |
  | ----------------------|-------------|
  | `:input`              | The request input to be sent
  | `:server`             | The remote server endpoint or a function (eg: \"http://localhost:8080\"); default is first entry from api definition
  | `:headers`            | A map of additional headers to be sent
  | `:options`            | A map of additional options to add to the request
  | `:as`                 | The format for content negotation; accepts `edn` or `json`; default: `json`
  | `:interceptors`       | The interceptor configuration to apply
  | `:throw-exceptions`   | Control throwing of exceptions in case of errors; default: `true`
  | `:decode-on-error`    | Control decoding of response body in case of errors; default: `true`
  "
  [{:keys [parsed-api command-defs] :as client}
   command
   {:keys [input
           server
           headers
           options
           as
           throw-exceptions
           decode-on-error
           interceptors]
    :or   {input            nil
           server           (get client :server)
           headers          {}
           as               :json
           throw-exceptions true
           decode-on-error  true
           interceptors     {}}
    :as   opts}]

  (ex/assert-spec-valid ::client client)
  (ex/assert-spec-valid ::command command)
  (ex/assert-spec-valid ::options opts)
  (ex/assert-spec-valid ::interceptors interceptors)
  ;;ensure command is valid
  (ex/assert-spec-valid (into #{} (keys command-defs)) command {:message "Invalid command"})

  (let [command-def (get command-defs command)
        {::reg/keys [commands]} parsed-api
        {:keys [input-spec
                path-spec
                params-spec
                response-spec
                input?
                params?
                pathelems?]} command-def
        command-map (get commands command)
        request-map (url/cmd->request command-map input)
        format      (cond
                      (= as :edn) "application/edn"
                      :else "application/json")]

    ;; validate command specifics
    (if input? (ex/assert-spec-valid input-spec (:body request-map)) {:message "Invalid input"})
    (if pathelems? (ex/assert-spec-valid path-spec input {:message "Invalid input path params"}))
    (if params? (ex/assert-spec-valid params-spec (:query-params request-map) {:message "Invalid input params"}))

    (let [target-url        (str (resolve-server server opts) (:url request-map))
          headers           (cond-> (assoc headers "Accept" format)
                              input? (assoc "Content-Type" format))
          base-req          (-> request-map
                              (merge default-options options)
                              (assoc :url target-url
                                     :headers headers
                                     :throw-exceptions throw-exceptions
                                     :decode-on-error decode-on-error
                                     :pool (:connection-pool client)))
          http-req          (cond-> base-req
                              input? (assoc :body (m/encode format input))
                              params? (assoc :query-params (:query-params request-map)))

          interceptor-chain (merge interceptors command-def)
          interceptor-chain (build-chain interceptor-chain)]

      (ixm/execute {:request http-req} interceptor-chain))))


;;;;
;; Specs
;;;;

;; api-definition


(s/def ::api-definition ::core/definition)

;; client
(s/def ::client #(instance? Client %))

;; invoke arguments
(s/def ::command keyword?)

(s/def :client-options/ssl-context any?)
(s/def :client-options/server (s/or :str string? :ifn ifn?))
(s/def ::client-options (s/nilable (s/keys :opt-un [:client-options/ssl-context
                                                    :client-options/server])))

;; config
(s/def ::input map?)
(s/def ::server string?)
(s/def ::as #{:json :edn})
(s/def ::throw-exceptions boolean?)
(s/def ::user-interceptors map?)
(s/def ::interceptors (s/keys :opt [::bpi/additional ::bpi/disabled]))
(s/def ::options (s/nilable (s/keys :opt-un [::input ::server ::headers
                                             ::as ::throw-exceptions
                                             ::interceptors])))

==> blueprint-client-0.2.22.jar: blueprint/client.clj <==
(ns blueprint.client
  (:require [blueprint.client.url :as url]
            [blueprint.core :as core]
            [blueprint.registry :as reg]
            [blueprint.interceptor :as bpi]
            [blueprint.client.interceptors :as interceptors]
            [aleph.http :as http]
            [muuntaja.core :as m]
            [exoscale.interceptor.manifold :as ixm]
            [blueprint.client.spec-gen :as bsg]
            [clojure.spec.alpha :as s]
            [exoscale.ex :as ex]))

(def default-interceptors
  [interceptors/execute-request
   interceptors/lens-response
   interceptors/decode
   interceptors/parse
   interceptors/get-response])

(defn- build-chain
  "Build an interceptor chain in three phases:

  - Start from the default chain
  - Process additional interceptors provided in `::additional`
  - Removes interceptors by name based on those provided in `::disabled`
  - Call the interceptor build step if any with the provided config"
  [config]
  (ex/assert-spec-valid ::interceptors config)
  (bpi/build-chain default-interceptors config))

(def build-with bpi/build-with)

;;
;; Http Client
;;

(defn- base-url
  "Gets the base url for the defined api."
  [{:keys [servers] :as parsed-api}]
  (if-let [prefix (get-in parsed-api [:blueprint.options :path-prefix])]
    (str (:url (first servers)) prefix)
    (:url (first servers))))

(defn- resolve-server
  "Resolves the correct server according to the configured option."
  [str-or-ifn  opts]
  (let [[t val] (s/conform :client-options/server str-or-ifn)]
    (cond (= t :str) val
          (= t :ifn) (val opts)
          :else (throw (ex-info (str "Unrecognized server option: " str-or-ifn) {:type t
                                                                                 :val  val})))))

(defrecord Client [parsed-api command-defs server connection-pool])

(defn make-client
  "Create a blueprint client from a blueprint api definition."
  ([api-def] (make-client api-def {}))
  ([api-def {:keys [ssl-context server] :as default-options}]

   (ex/assert-spec-valid ::api-definition api-def)
   (ex/assert-spec-valid ::client-options default-options)

   (let [parsed-api      (core/parse api-def)
         command-defs    (bsg/apidef->commands-map parsed-api)
         connection-pool (if ssl-context
                           (http/connection-pool
                            {:connection-options {:ssl-context ssl-context}})
                           http/default-connection-pool)
         server          (or server (base-url parsed-api))]
     (->Client parsed-api command-defs server connection-pool))))

(def default-options
  {:pool-timeout       10e3
   :connection-timeout 10e3
   :request-timeout    10e3
   :read-timeout       10e3})

(defn invoke
  "Invoke a `command` on the remote blueprint-based api.
  Clients are created with `(blueprint.client/make-client api-def)`.
  Supported options in `opts`:

  | key                   | description |
  | ----------------------|-------------|
  | `:input`              | The request input to be sent
  | `:server`             | The remote server endpoint or a function (eg: \"http://localhost:8080\"); default is first entry from api definition
  | `:headers`            | A map of additional headers to be sent
  | `:options`            | A map of additional options to add to the request
  | `:as`                 | The format for content negotation; accepts `edn` or `json`; default: `json`
  | `:interceptors`       | The interceptor configuration to apply
  | `:throw-exceptions`   | Control throwing of exceptions in case of errors; default: `true`
  | `:decode-on-error`    | Control decoding of response body in case of errors; default: `true`
  "
  [{:keys [parsed-api command-defs] :as client}
   command
   {:keys [input
           server
           headers
           options
           as
           throw-exceptions
           decode-on-error
           interceptors]
    :or   {input            nil
           server           (get client :server)
           headers          {}
           as               :json
           throw-exceptions true
           decode-on-error  true
           interceptors     {}}
    :as   opts}]

  (ex/assert-spec-valid ::client client)
  (ex/assert-spec-valid ::command command)
  (ex/assert-spec-valid ::options opts)
  (ex/assert-spec-valid ::interceptors interceptors)
  ;;ensure command is valid
  (ex/assert-spec-valid (into #{} (keys command-defs)) command {:message "Invalid command"})

  (let [command-def (get command-defs command)
        {::reg/keys [commands ns]} parsed-api
        {:as request-map :keys [body query-params]} (url/cmd->request command-def input)
        format      (if (= as :edn)
                      "application/edn"
                      "application/json")]

    ;; validate command specifics
    (ex/assert-spec-valid (core/command-input-spec ns command)
                          (:body request-map)
                          {:message "Invalid input"})

    (ex/assert-spec-valid (core/command-path-spec ns command)
                          input
                          {:message "Invalid input path params"})

    (ex/assert-spec-valid (core/command-params-spec ns command)
                          (:query-params request-map)
                          {:message "Invalid input params"})

    (let [target-url        (str (resolve-server server opts) (:url request-map))
          headers           (cond-> (assoc headers "Accept" format)
                              body (assoc "Content-Type" format))
          base-req          (-> request-map
                                (merge default-options options)
                                (assoc :url target-url
                                       :headers headers
                                       :throw-exceptions throw-exceptions
                                       :decode-on-error decode-on-error
                                       :pool (:connection-pool client)))
          http-req          (cond-> base-req
                              body (assoc :body (m/encode format input))
                              query-params (assoc :query-params (:query-params request-map)))

          interceptor-chain (assoc interceptors :command command-def)
          interceptor-chain (build-chain interceptor-chain)]

      (ixm/execute {:request http-req} interceptor-chain))))


;;;;
;; Specs
;;;;

;; api-definition


(s/def ::api-definition :blueprint.core/definition)

;; client
(s/def ::client #(instance? Client %))

;; invoke arguments
(s/def ::command keyword?)

(s/def :client-options/ssl-context any?)
(s/def :client-options/server (s/or :str string? :ifn ifn?))
(s/def ::client-options (s/nilable (s/keys :opt-un [:client-options/ssl-context
                                                    :client-options/server])))

;; config
(s/def ::input map?)
(s/def ::server string?)
(s/def ::as #{:json :edn})
(s/def ::throw-exceptions boolean?)
(s/def ::user-interceptors map?)
(s/def ::interceptors (s/keys :opt [::bpi/additional ::bpi/disabled]))
(s/def ::options (s/nilable (s/keys :opt-un [::input ::server ::headers
                                             ::as ::throw-exceptions
                                             ::interceptors])))
